import logging

from flask import jsonify
from flask_limiter import Limiter
from flask_limiter.util import get_remote_address


def rate_limit_error_handler(_):
    """Handler for rate limit errors"""
    logging.warning("Rate limit exceeded")
    return jsonify(
        {
            "error": "Too many requests",
            "message": "Allowed number or requests reached. Try again later.",
        }
    ), 429


def init_limiter(app):
    """Initialize the rate limiter"""
    limiter = Limiter(
        app=app,
        key_func=get_remote_address,
        default_limits=app.config.get(
            "RATELIMIT_DEFAULT", "150000 per day;6000 per hour"
        ).split(";"),
        storage_uri=app.config.get("RATELIMIT_STORAGE_URI", "memory://"),
    )

    app.register_error_handler(429, rate_limit_error_handler)
    return limiter



import re

from flask import Blueprint, jsonify
from pydantic.v1 import (
    BaseModel,
)

DEFAULT_RESPONSES = {
    "200": {"description": "Success"},
    "400": {"description": "Invalid input data"},
    "404": {"description": "Invalid or non existing object"},
    "500": {"description": "Error calling method on adapter"},
}


def to_openapi_path(route: str) -> str:
    """
    Adds the "{" "}" to the route arguments that OpenAPI requires

    Args:
        route(str): The route

    Returns:
        returns: the openapi path, with the {} srounding the arguments
    """
    _s = re.sub(r"<[^:]+:", "{", route)

    # Replace closing ">" with "}"
    return _s.replace(">", "}")


class OpenAPISpec:
    def __init__(self, name: str, url_prefix, api_version: str, api_title: str):
        self.bp = Blueprint(name, __name__, url_prefix=url_prefix)

        self._openapi_spec = {
            "openapi": "3.0.0",
            "info": {"title": api_title, "version": api_version},
            "paths": {},
            "components": {"schemas": {}},
        }

        self.bp.add_url_rule(
            "/openapi.json", "openapi", self._serve_openapi, methods=["GET"]
        )
        self.bp.add_url_rule("/docs", "redoc_ui", self._serve_redoc_ui, methods=["GET"])
        self.bp.add_url_rule(
            "/docs_redoc", "redoc_ui", self._serve_redoc_ui, methods=["GET"]
        )
        self.bp.add_url_rule(
            "/docs_swagger", "swagger_ui", self._serve_swagger_ui, methods=["GET"]
        )
        self.bp.add_url_rule(
            "/docs_elements", "elements_ui", self._serve_elements_ui, methods=["GET"]
        )

    def add_openapi_path(
        self,
        prefix: str,
        route: str,
        export: dict[str, str],
        http_method: str,
        view_func,
    ):
        """Adds API endpoints to OpenAPI spec."""
        open_api_path = to_openapi_path(prefix + route)

        self._openapi_spec["paths"].setdefault(open_api_path, {})[
            http_method.lower()
        ] = {
            "summary": f"{http_method} {export['attr']}",
            "description": str(view_func.__doc__),
            "tags": [prefix],
            "parameters": [
                {
                    "name": "object_id",
                    "in": "path",
                    "required": True,
                    "schema": {"type": "string"},
                }
            ],
            "requestBody": {},
            "responses": dict(DEFAULT_RESPONSES),
        }

    def add_openapi_schema(
        self, prefix: str, route: str, http_method: str, model: type[BaseModel]
    ):
        """Adds Pydantic model definitions to OpenAPI schema."""

        open_api_path = to_openapi_path(prefix + route)
        schema_name = model.__name__

        if schema_name not in self._openapi_spec["components"]["schemas"]:
            self._openapi_spec["components"]["schemas"][schema_name] = model.schema()

        self._openapi_spec["paths"][open_api_path][http_method.lower()][
            "requestBody"
        ].update(
            {
                "description": model.__name__,
                "required": True,
                "content": {
                    "application/json": {
                        "schema": {"$ref": f"#/components/schemas/{model.__name__}"}
                    }
                },
            }
        )

    def add_openapi_response(
        self, prefix: str, route: str, http_method: str, model: type[BaseModel]
    ):
        """Adds response schema to OpenAPI documentation."""

        open_api_path = to_openapi_path(prefix + route)
        schema_name = model.__name__ if isinstance(model, BaseModel) else None

        if schema_name:
            if schema_name not in self._openapi_spec["components"]["schemas"]:
                self._openapi_spec["components"]["schemas"][schema_name] = (
                    model.schema()
                )

            self._openapi_spec["paths"][open_api_path][http_method.lower()][
                "responses"
            ]["200"] = {
                "description": "Successful response",
                "content": {
                    "application/json": {
                        "schema": {"$ref": f"#/components/schemas/{model.__name__}"}
                    }
                },
            }

    def _serve_openapi(self):
        return jsonify(self._openapi_spec)

    def _serve_redoc_ui(self):
        """Serves the ReDoc UI for OpenAPI documentation."""
        return """
        <!DOCTYPE html>
        <html>
        <head>
            <title>Redoc</title>
            <!-- needed for adaptive design -->
            <meta charset="utf-8"/>
            <meta name="viewport" content="width=device-width, initial-scale=1">
            <link href="https://fonts.googleapis.com/css?family=Montserrat:300,400,700|Roboto:300,400,700" rel="stylesheet">

            <!--
            Redoc doesn't change outer page styles
            -->
            <style>
            body {
                margin: 0;
                padding: 0;
            }
            </style>
        </head>
        <body>
            <redoc spec-url='openapi.json'></redoc>
            <script src="https://cdn.redoc.ly/redoc/latest/bundles/redoc.standalone.js"> </script>
        </body>
        </html>
        """  # noqa: E501

    def _serve_swagger_ui(self):
        """Serves the Swagger UI for OpenAPI documentation."""
        return """
        <!DOCTYPE html>
        <html lang="en">
        <head>
        <meta charset="utf-8" />
        <meta name="viewport" content="width=device-width, initial-scale=1" />
        <meta name="description" content="SwaggerUI" />
        <title>SwaggerUI</title>
        <link rel="stylesheet" href="https://unpkg.com/swagger-ui-dist@5.11.0/swagger-ui.css" />
        </head>
        <body>
        <div id="swagger-ui"></div>
        <script src="https://unpkg.com/swagger-ui-dist@5.11.0/swagger-ui-bundle.js" crossorigin></script>
        <script>
        window.onload = () => {
            window.ui = SwaggerUIBundle({
            url: 'openapi.json',
            dom_id: '#swagger-ui',
            });
        };
        </script>
        </body>
        </html>
        """  # noqa: E501

    def _serve_elements_ui(self):
        """Serves the Elements UI for OpenAPI documentation."""
        return """
            <!doctype html>
            <html lang="en">
            <head>
                <meta charset="utf-8">
                <meta name="viewport" content="width=device-width, initial-scale=1, shrink-to-fit=no">
                <title>Elements in HTML</title>

                <script src="https://unpkg.com/@stoplight/elements/web-components.min.js"></script>
                <link rel="stylesheet" href="https://unpkg.com/@stoplight/elements/styles.min.css">
            </head>
            <body>

                <elements-api
                apiDescriptionUrl="openapi.json"
                router="hash"
                />

            </body>
            </html>
        """  # noqa: E501



import logging

from flask import (
    Blueprint,
    Response,
    jsonify,
)


def init_route(app, server, url_prefix):
    bp = Blueprint("beamline", __name__, url_prefix=url_prefix)

    @bp.route("/", methods=["GET"])
    @server.restrict
    def beamline_get_all_attributes():
        return jsonify(app.beamline.beamline_get_all_attributes())

    @bp.route("/prepare_beamline", methods=["PUT"])
    @server.require_control
    @server.restrict
    def prepare_beamline_for_sample():
        """
        Prepare the beamline for a new sample.
        """
        try:
            app.beamline.prepare_beamline_for_sample()
        except Exception:
            msg = "Cannot prepare the Beamline for a new sample"
            logging.getLogger("HWR").exception(msg)
            return Response(status=200)
        return Response(status=200)

    return bp



import logging
import math
import re

from flask_login import current_user
from mxcubecore import HardwareRepository as HWR
from mxcubecore.model import queue_model_objects as qmo
from mxcubecore.model.lims_session import LimsSessionManager

from mxcubeweb.core.components.component_base import ComponentBase

VALID_SAMPLE_NAME_REGEXP = re.compile("^[a-zA-Z0-9:+_-]+$")


class NoSessionError(Exception):
    """Exception raised when no session is selected in LIMS."""


class Lims(ComponentBase):
    def __init__(self, app, config):
        super().__init__(app, config)

    def new_sample_list(self):
        return {"sampleList": {}, "sampleOrder": []}

    def init_sample_list(self):
        self.sample_list_set(self.new_sample_list())

    def sample_list_set(self, sample_list):
        self.app.SAMPLE_LIST = sample_list

    def sample_list_set_order(self, sample_order):
        self.app.SAMPLE_LIST["sampleOrder"] = sample_order

    def sample_list_get(self, loc=None, current_queue=None):
        self.synch_sample_list_with_queue(current_queue)
        res = self.app.SAMPLE_LIST

        if loc:
            res = self.app.SAMPLE_LIST.get("sampleList").get(loc, {})

        return res

    def sample_list_sync_sample(self, lims_sample):
        lims_code = lims_sample.get("code", None)
        lims_location = lims_sample.get("lims_location")
        sample_to_update = None

        # LIMS sample has code, check if the code was read by SC
        if lims_code and self.app.sample_changer.sc_contents_from_code_get(lims_code):
            sample_to_update = self.app.sample_changer.sc_contents_from_code_get(
                lims_code
            )
        elif lims_location:
            # Asume that the samples have been put in the right place of the SC
            sample_to_update = self.app.sample_changer.sc_contents_from_location_get(
                lims_location
            )

        if sample_to_update:
            loc = sample_to_update["sampleID"]
            self.sample_list_update_sample(loc, lims_sample)

    def synch_sample_list_with_queue(self, current_queue=None):
        if not current_queue:
            current_queue = self.app.queue.queue_to_dict(include_lims_data=True)

        current_queue.get("sample_order", [])

        for loc, data in self.app.SAMPLE_LIST["sampleList"].items():
            if loc in current_queue:
                sample = current_queue[loc]

                # Don't synchronize, lims attributes from queue sample, if
                # they are already set by sc or lims
                if data.get("sampleName", ""):
                    sample.pop("sampleName")

                if data.get("proteinAcronym", ""):
                    sample.pop("proteinAcronym")

                # defaultSubDir and prefix are derived from proteinAcronym
                # and/or sampleName so make sure that those are removed from
                # queue sample so that they can be updated if changed.
                if data.get("proteinAcronym", "") or data.get("sampleName", ""):
                    sample.pop("defaultPrefix")
                    sample.pop("defaultSubDir")

                # Make sure that sample in queue is updated with lims information
                model, entry = self.app.queue.get_entry(sample["queueID"])
                model.set_from_dict(data)

                # Update sample location, location is Manual for free pin mode
                # in MXCuBE Web
                model.loc_str = data.get("sampleID", -1)
                model.free_pin_mode = data.get("location", "") == "Manual"

                self.sample_list_update_sample(loc, sample)

    def sample_list_update_sample(self, loc, sample):
        _sample = self.app.SAMPLE_LIST["sampleList"].get(loc, {})

        # If sample exists in sample list update it, otherwise add it
        if _sample:
            _sample.update(sample)
        else:
            self.app.SAMPLE_LIST["sampleList"][loc] = sample
            self.app.SAMPLE_LIST["sampleOrder"].append(loc)

        return self.app.SAMPLE_LIST["sampleList"].get(loc, {})

    def apply_template(self, params, sample_model, path_template):
        # Apply subdir template if used:
        if "{" in params.get("subdir", ""):
            if sample_model.crystals[0].protein_acronym:
                params["subdir"] = params["subdir"].format(
                    NAME=sample_model.get_name(),
                    ACRONYM=sample_model.crystals[0].protein_acronym,
                )
            else:
                stripped = params["subdir"][0 : params["subdir"].find("{")]
                params["subdir"] = stripped + sample_model.get_name()

            # The template was only applied partially if subdir ends with '-'
            # probably because either acronym or protein name is null in LIMS
            if params["subdir"].endswith("-"):
                params["subdir"] = sample_model.get_name()

        # Making sure that there are no ":" left from the sample name incase
        # no synchronisation with LIMS was done
        params["subdir"] = params["subdir"].replace(":", "-")

        if "{" in params.get("prefix", ""):
            sample = self.app.SAMPLE_LIST["sampleList"].get(sample_model.loc_str, {})
            prefix = self.get_default_prefix(sample)
            shape = params["shape"] if params["shape"] > 0 else ""
            params["prefix"] = params["prefix"].format(PREFIX=prefix, POSITION=shape)

            if params["prefix"].endswith("_"):
                params["prefix"] = params["prefix"][:-1]

        # mxcube web passes entire prefix as prefix, including reference, mad and wedge
        # prefix. So we strip those before setting the actual base_prefix.
        params["prefix"] = self.strip_prefix(path_template, params["prefix"])

    def strip_prefix(self, pt, prefix):
        """
        Strips the reference, wedge and mad prefix from a given prefix. For example
        removes ref- from the beginning and _w[n] and -pk, -ip, -ipp from the end.

        :param PathTemplate pt: path template used to create the prefix
        :param str prefix: prefix from the client
        :returns: stripped prefix
        """
        if (
            pt.reference_image_prefix
            and pt.reference_image_prefix == prefix[0 : len(pt.reference_image_prefix)]
        ):
            prefix = prefix[len(pt.reference_image_prefix) + 1 :]

        if pt.wedge_prefix and pt.wedge_prefix == prefix[-len(pt.wedge_prefix) :]:
            prefix = prefix[: -(len(pt.wedge_prefix) + 1)]

        if pt.mad_prefix and pt.mad_prefix == prefix[-len(pt.mad_prefix) :]:
            prefix = prefix[: -(len(pt.mad_prefix) + 1)]

        return prefix

    def get_session_manager(self) -> LimsSessionManager:
        return HWR.beamline.lims.session_manager

    def is_rescheduled_session(self, session):
        """
        Returns true is the session is rescheduled. That means that either currently is not the expected timeslot
        or because it is not in the expected beamline
        """
        return not (session.is_scheduled_beamline and session.is_scheduled_time)

    def allow_session(self, session):
        HWR.beamline.lims.allow_session(session)

    def select_session(self, session_id: str) -> bool:
        """
        param session_id : this is a identifier that could be proposal name or session_id depending of the type of LIMS login type
        """
        logging.getLogger("MX3.HWR").debug("select_session session_id=%s" % session_id)

        # Selecting the active session in the LIMS object
        try:
            session = HWR.beamline.lims.set_active_session_by_id(session_id)
        except Exception as exc:
            logging.getLogger("MX3.HWR").exception(
                "No session candidate. Force signout."
            )
            self.app.usermanager.signout()
            raise NoSessionError from exc

        if (
            HWR.beamline.lims.is_user_login_type()
            and "Commissioning" in session.title
            and hasattr(HWR.beamline.session, "set_in_commissioning")
        ):
            HWR.beamline.session.set_in_commissioning(self.get_proposal_info())
            logging.getLogger("MX3.HWR").info("[LIMS] Commissioning proposal flag set.")

        if HWR.beamline.session.session_id != HWR.beamline.lims.get_session_id():
            # ruff: noqa: G004
            logging.getLogger("MX3.HWR").info(
                f"[LIMS] New session, clearing queue and sample list for {session.code}{session.number}"
            )

            # Clear data collection queue (HardwareObject)
            self.app.queue.clear_queue()

            # Remove any items on the sample view (shapes)
            HWR.beamline.sample_view.clear_all()

            # Re-initialize the samplelist
            self.app.lims.init_sample_list()

            # Get sample list and send update to client
            self.app.sample_changer.get_sample_list()
            self.app.server.emit("update_queue", {}, namespace="/hwr")

            HWR.beamline.session.proposal_code = session.code
            HWR.beamline.session.proposal_number = session.number
            HWR.beamline.session.session_id = HWR.beamline.lims.get_session_id()
            HWR.beamline.session.proposal_id = session.proposal_id
            HWR.beamline.session.set_session_start_date(session.start_date)

        logging.getLogger("MX3.HWR").info(
            "[LIMS] Selected session. proposal=%s session_id=%s.",
            session.proposal_name,
            session.session_id,
        )

        if self.is_rescheduled_session(session):
            logging.getLogger("MX3.HWR").info(
                "[LIMS] Session is rescheduled in time or beamline."
            )
            self.allow_session(session)

        HWR.beamline.session.prepare_directories(session)

        # save selected proposal in users db
        current_user.selected_proposal = session.session_id
        self.app.usermanager.update_user(current_user)

        logging.getLogger("user_log").info(
            "[LIMS] Proposal selected session_id=%s.", session_id
        )

        return True

    def get_default_prefix(self, sample_data, generic_name=False):
        if isinstance(sample_data, dict):
            sample = qmo.Sample()
            sample.code = sample_data.get("code", "")
            sample.name = sample_data.get("sampleName", "").replace(":", "-")
            sample.location = sample_data.get("location", "").split(":")
            sample.lims_id = sample_data.get("limsID", -1)
            sample.crystals[0].protein_acronym = sample_data.get("proteinAcronym", "")
        else:
            sample = sample_data

        return HWR.beamline.session.get_default_prefix(sample, generic_name)

    def get_default_subdir(self, sample_data):
        return HWR.beamline.session.get_default_subdir(sample_data)

    def synch_with_lims(self, lims_name):
        self.app.queue.queue_clear()
        self.app.sample_changer.get_sample_list()

        samples_info_list = HWR.beamline.lims.get_samples(lims_name)
        for sample_info in samples_info_list:
            sample_info["limsID"] = sample_info.pop("sampleId")
            sample_info["defaultPrefix"] = self.get_default_prefix(sample_info)
            sample_info["defaultSubDir"] = self.get_default_subdir(sample_info)

            if not VALID_SAMPLE_NAME_REGEXP.match(sample_info["sampleName"]):
                raise AttributeError(
                    "sample name for sample %s contains an incorrect character"
                    % sample_info
                )

            try:
                basket = int(sample_info["containerSampleChangerLocation"])
            except (TypeError, ValueError, KeyError):
                continue
            else:
                if HWR.beamline.sample_changer.__class__.__TYPE__ in [
                    "Flex Sample Changer",
                    "FlexHCD",
                    "RoboDiff",
                ]:
                    cell = math.ceil((basket) / 3.0)
                    puck = basket - 3 * (cell - 1)
                    sample_info["containerSampleChangerLocation"] = "%d:%d" % (
                        cell,
                        puck,
                    )

            try:
                lims_location = sample_info[
                    "containerSampleChangerLocation"
                ] + ":%02d" % int(sample_info["sampleLocation"])
            except Exception:
                logging.getLogger("MX3.HWR").info(
                    "[LIMS] Could not parse sample loaction from"
                    " LIMS, (perhaps not set ?)"
                )
            else:
                sample_info["lims_location"] = lims_location

                self.sample_list_sync_sample(sample_info)

        return self.sample_list_get()

    def get_lims_samples(self, lims_name: str) -> dict:
        """
        Gets samples from LIMS and filters to include only LIMS-linked entries.

        This method synchronizes the sample list with the specified LIMS,
        filters for entries that have a `limsID`, and sets the new sample list
        accordingly.

        Args: Name of the LIMS system to synchronize with.

        Returns: The updated sample list with only LIMS samples.
        """
        self.synch_with_lims(lims_name)
        new_sample_list = {"sampleList": {}, "sampleOrder": []}

        try:
            for loc, data in self.app.SAMPLE_LIST.get("sampleList", {}).items():
                if data.get("limsID"):
                    new_sample_list["sampleList"][loc] = data
                    new_sample_list["sampleOrder"].append(loc)
        except Exception:
            logging.getLogger("MX3.HWR").exception(
                "Error while filtering LIMS samples for '%s':", lims_name
            )
            return {"sampleList": {}, "sampleOrder": []}

        self.sample_list_set(new_sample_list)
        return self.app.SAMPLE_LIST



class CSPMiddleware:
    """
    Middleware to add Content Security Policy headers to responses.
    """

    def __init__(self, app, config):
        self.app = app
        self.config = config
        self.enabled = config.get("CSP_ENABLED", True)
        self.policy = config.get("CSP_POLICY", {})
        self.report_only = config.get("CSP_REPORT_ONLY", False)
        self.report_uri = config.get("CSP_REPORT_URI", "")

    def __call__(self, environ, start_response):
        if not self.enabled:
            return self.app(environ, start_response)

        def _start_response(status, headers, exc_info=None):
            if self.enabled:
                policy_str = self._build_policy_string()
                header_name = (
                    "Content-Security-Policy-Report-Only"
                    if self.report_only
                    else "Content-Security-Policy"
                )
                headers.append((header_name, policy_str))

            return start_response(status, headers, exc_info)

        return self.app(environ, _start_response)

    def _build_policy_string(self) -> str:
        # CSP keywords that need to be quoted
        csp_keywords = {
            "self",
            "unsafe-inline",
            "unsafe-eval",
            "none",
            "strict-dynamic",
            "unsafe-hashes",
            "report-sample",
            "wasm-unsafe-eval",
            "script",
        }

        def quote_if_keyword(source):
            if source in csp_keywords:
                return f"'{source}'"
            return source

        parts = []
        for directive, sources in self.policy.items():
            if sources:
                quoted_sources = [quote_if_keyword(s) for s in sources]
                parts.append(f"{directive} {' '.join(quoted_sources)}")

        if self.report_uri:
            parts.append(f"report-uri {self.report_uri}")

        return "; ".join(parts)



from flask import (
    Blueprint,
    Response,
    jsonify,
    request,
)


def init_route(app, server, url_prefix):
    bp = Blueprint("workflow", __name__, url_prefix=url_prefix)

    @bp.route("/", methods=["GET"])
    @server.restrict
    def workflow():
        return jsonify(app.workflow.get_available_workflows())

    @bp.route("/", methods=["POST"])
    @server.restrict
    def submit_parameters():
        data = request.get_json()
        app.workflow.submit_parameters(data)
        return Response(status=200)

    @bp.route("/gphl", methods=["POST"])
    @server.restrict
    def submit_gphl_parameters():
        data = request.get_json()
        app.workflow.update_gphl_parameters(data)
        return Response(status=200)

    return bp
